import os
import cv2
import argparse
from pyagender.pyagender import PyAgender
from pyagender import VERSION
from keras.utils.data_utils import get_file

# 64x64 (RGB, padded) IMDB dataset trained for 28 epochs
pretrained_model = "https://github.com/yu4u/age-gender-estimation/releases/download/v0.5/weights.28-3.73.hdf5"
modhash = 'fbe63257a054c1c5466cfd7bf14646d6'
pretrained_model_filename = "weights.28-3.73.hdf5"


def get_args():
    parser = argparse.ArgumentParser(
        description="Detect and get faces age-gender from a picture file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("IMAGE", type=str, nargs='?', default='', help="path to image file (jpg, png, bmp)")
    parser.add_argument("-v", default=False, action='store_true', help="version info")
    args = parser.parse_args()
    return args


def main():
    args = get_args()

    if args.v:
        print(f'{VERSION}, model: {pretrained_model_filename}')
        exit()

    weight_file = get_file(pretrained_model_filename, pretrained_model,
                           cache_subdir="pretrained_models",
                           file_hash=modhash,
                           cache_dir=os.path.dirname(os.path.abspath(__file__)))
    if not args.IMAGE:
        exit(0)

    pyagender = PyAgender(resnet_weights=weight_file)

    # trained model colorspace seems like BGR https://github.com/yu4u/age-gender-estimation/issues/51
    image = cv2.imread(args.IMAGE)
    faces = pyagender.detect_genders_ages(image)

    if len(faces) < 1:
        exit(0)

    for face in faces:
        print(face)

    # ZEU


if __name__ == '__main__':
    main()



import setuptools
import pyagender

REQUIRED_PACKAGES = [
    'numpy >= 1.12',
    'Keras >= 2.1',
    'TensorFlow >= 1.10, < 1.19',
    'opencv-python >= 3.3.0+contrib, < 3.99'
]

CONSOLE_SCRIPTS = [
    'py-agender = pyagender.pyagender_cli:main'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py-agender",
    version=pyagender.VERSION,
    author="Michael Butlitsky",
    author_email="aristofun@yandex.ru",
    description="Simple opencv & tensorflow based solution to estimate Faces, Age and Gender on pictures",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aristofun/py-agender",
    packages=setuptools.find_packages(),
    classifiers=(
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        "Environment :: Console",
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'License :: Other/Proprietary License',
        "Operating System :: OS Independent",
    ),
    zip_safe=False,
    install_requires=REQUIRED_PACKAGES,
    entry_points={'console_scripts': CONSOLE_SCRIPTS})



import os
import sys
import cv2
import numpy as np


sys.setrecursionlimit(2 ** 20)
# np.random.seed(2 ** 10)

from pyagender.wide_resnet import WideResNet


class PyAgender():
    """
    Age-Gender estimator with built-in OpenCV face detector.

    Based on https://github.com/asmith26/wide_resnets_keras ResNet
    and pre trained model from https://github.com/yu4u/age-gender-estimation
    """

    def __init__(self,
                 haar_cascade='pretrained_models/haarcascade_frontalface_default.xml',
                 cv_scaleFactor=1.1,
                 cv_minNeighbors=8,
                 cv_minSize=(64, 64),
                 cv_flags=cv2.CASCADE_SCALE_IMAGE,
                 resnet_weights='pretrained_models/weights.28-3.73.hdf5',
                 resnet_imagesize=64):
        self.cv_face_detector = cv2.CascadeClassifier(
            os.path.join(os.path.dirname(__file__), haar_cascade))
        self.cv_scaleFactor = cv_scaleFactor
        self.cv_minNeighbors = cv_minNeighbors
        self.cv_minSize = cv_minSize
        self.cv_flags = cv_flags
        self.resnet_imagesize = resnet_imagesize
        self.resnet = WideResNet(image_size=resnet_imagesize)()
        self.resnet.load_weights(os.path.join(os.path.dirname(__file__), resnet_weights))

    def detect_genders_ages(self, image):
        """
        Detects all faces using OpenCV Haar cascades options provided in the constructor

        :param image: image to detect
        :return: array of dicts or empty array (no detections)
                 {left: 34, top: 11, right: 122, bottom: 232, width:(r-l), height: (b-t),
                  gender: 0.67, age: 23.5}
                  
                 gender > 0.5 == female
        """
        faceregions = self.detect_faces(image, margin=0.4)

        for face in faceregions:
            face['gender'], face['age'] = self.gender_age(image,
                                                          left=face['left'], top=face['top'],
                                                          width=face['width'],
                                                          height=face['height'])

        return faceregions

    def gender_age(self, face_image, left=0, top=0, width=None, height=None):
        """
        Assuming ready to use face region on input (no detection to do)
        
        :param face_image: CV image object to feed into CNN age-gender estimator
        :param left,top left upper corner of a face in the image
        :param w,h width,height of a face (face_image == the whole face by default)

        :return: [gender, age] evaluation (gender > 0.5 == female, age is float)
        """
        img_h, img_w, __ = np.shape(face_image)

        if width is not None:
            img_w = width
        if height is not None:
            img_h = height

        # Crop & resize image to 64pix box
        test_img = PyAgender.aspect_resize(face_image[top:top + img_h, left:left + img_w],
                                           self.resnet_imagesize, self.resnet_imagesize)

        # predict ages and genders of the detected faces
        result = self.resnet.predict(np.array([test_img]))
        predicted_genders = result[0]
        ages = np.arange(0, 101).reshape(101, 1)
        predicted_ages = result[1].dot(ages).flatten()
        # print(f'gender: {predicted_genders}')
        # print(f'predicted_ages: {predicted_ages}')

        return predicted_genders[0][0], predicted_ages[0]

    def detect_faces(self, image, margin=0.2):
        """
        :param image: Original image (in opencv BGR) to find faces on
        :param padding: additional margin widht/height percentage
        :return:
            array of face image rectangles {left: 34, top: 11, right: 122, bottom: 232, width:(r-l), height: (b-t)}
        """
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        img_h, img_w = np.shape(gray)

        face_results = []

        faces = self.cv_face_detector.detectMultiScale(gray,
                                                       scaleFactor=self.cv_scaleFactor,
                                                       minNeighbors=self.cv_minNeighbors,
                                                       minSize=self.cv_minSize,
                                                       flags=self.cv_flags)
        for (x, y, w, h) in faces:
            xi1 = max(int(x - margin * w), 0)
            xi2 = min(int(x + w + margin * w), img_w - 1)
            yi1 = max(int(y - margin * h), 0)
            yi2 = min(int(y + h + margin * h), img_h - 1)
            detection = {'left': xi1, 'top': yi1, 'right': xi2, 'bottom': yi2,
                         'width': (xi2 - xi1), 'height': (yi2 - yi1)}
            face_results.append(detection)

        return face_results

    @staticmethod
    def aspect_resize(image, width, height, padding=cv2.BORDER_REPLICATE, color=[0, 0, 0]):
        """
        Letterboxing image resize preserving original aspect ratio, padding if necessary
        :param image: cv2 compatible (x,y,3) shape image data
        :param width: desired width in pixels
        :param height: desired height in pixels
        :param padding: OpenCV padding strategy (see https://docs.opencv.org/3.1.0/d3/df2/tutorial_py_basic_ops.html)
        :param color: padding cv2 color for cv2.BORDER_CONSTANT padding strategy

        :return: resized copy of image in cv2 default format
        """
        old_size = image.shape[:2]
        # determine the longest side
        ratio = min(float(height) / old_size[0], float(width) / old_size[1])
        # resize accordingly
        new_size = tuple([int(x * ratio) for x in old_size])
        im = cv2.resize(image, (new_size[1], new_size[0]))

        delta_w = width - new_size[1]
        delta_h = height - new_size[0]

        top, bottom = delta_h // 2, delta_h - (delta_h // 2)
        left, right = delta_w // 2, delta_w - (delta_w // 2)

        new_im = cv2.copyMakeBorder(im, top, bottom, left, right, padding, value=color)
        return new_im



# This code is imported from the following project: https://github.com/asmith26/wide_resnets_keras

import logging
from keras.models import Model
from keras.layers import Input, Activation, add, Dense, Flatten, Dropout
from keras.layers.convolutional import Conv2D, AveragePooling2D
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l2
from keras import backend as K


class WideResNet:
    def __init__(self, image_size=64, depth=16, k=8):
        self._depth = depth
        self._k = k
        self._dropout_probability = 0
        self._weight_decay = 0.0005
        self._use_bias = False
        self._weight_init = "he_normal"

        if K.image_dim_ordering() == "th":
            logging.debug("image_dim_ordering = 'th'")
            self._channel_axis = 1
            self._input_shape = (3, image_size, image_size)
        else:
            logging.debug("image_dim_ordering = 'tf'")
            self._channel_axis = -1
            self._input_shape = (image_size, image_size, 3)

    # Wide residual network http://arxiv.org/abs/1605.07146
    def _wide_basic(self, n_input_plane, n_output_plane, stride):
        def f(net):
            # format of conv_params:
            #               [ [kernel_size=("kernel width", "kernel height"),
            #               strides="(stride_vertical,stride_horizontal)",
            #               padding="same" or "valid"] ]
            # B(3,3): orignal <<basic>> block
            conv_params = [[3, 3, stride, "same"],
                           [3, 3, (1, 1), "same"]]

            n_bottleneck_plane = n_output_plane

            # Residual block
            for i, v in enumerate(conv_params):
                if i == 0:
                    if n_input_plane != n_output_plane:
                        net = BatchNormalization(axis=self._channel_axis)(net)
                        net = Activation("relu")(net)
                        convs = net
                    else:
                        convs = BatchNormalization(axis=self._channel_axis)(net)
                        convs = Activation("relu")(convs)

                    convs = Conv2D(n_bottleneck_plane, kernel_size=(v[0], v[1]),
                                   strides=v[2],
                                   padding=v[3],
                                   kernel_initializer=self._weight_init,
                                   kernel_regularizer=l2(self._weight_decay),
                                   use_bias=self._use_bias)(convs)
                else:
                    convs = BatchNormalization(axis=self._channel_axis)(convs)
                    convs = Activation("relu")(convs)
                    if self._dropout_probability > 0:
                        convs = Dropout(self._dropout_probability)(convs)
                    convs = Conv2D(n_bottleneck_plane, kernel_size=(v[0], v[1]),
                                   strides=v[2],
                                   padding=v[3],
                                   kernel_initializer=self._weight_init,
                                   kernel_regularizer=l2(self._weight_decay),
                                   use_bias=self._use_bias)(convs)

            # Shortcut Connection: identity function or 1x1 convolutional
            #  (depends on difference between input & output shape - this
            #   corresponds to whether we are using the first block in each
            #   group; see _layer() ).
            if n_input_plane != n_output_plane:
                shortcut = Conv2D(n_output_plane, kernel_size=(1, 1),
                                  strides=stride,
                                  padding="same",
                                  kernel_initializer=self._weight_init,
                                  kernel_regularizer=l2(self._weight_decay),
                                  use_bias=self._use_bias)(net)
            else:
                shortcut = net

            return add([convs, shortcut])

        return f

    # "Stacking Residual Units on the same stage"
    def _layer(self, block, n_input_plane, n_output_plane, count, stride):
        def f(net):
            net = block(n_input_plane, n_output_plane, stride)(net)
            for i in range(2, int(count + 1)):
                net = block(n_output_plane, n_output_plane, stride=(1, 1))(net)
            return net

        return f

    #    def create_model(self):
    def __call__(self):
        logging.debug("Creating model...")

        assert ((self._depth - 4) % 6 == 0)
        n = (self._depth - 4) / 6

        inputs = Input(shape=self._input_shape)

        n_stages = [16, 16 * self._k, 32 * self._k, 64 * self._k]

        conv1 = Conv2D(filters=n_stages[0], kernel_size=(3, 3),
                       strides=(1, 1),
                       padding="same",
                       kernel_initializer=self._weight_init,
                       kernel_regularizer=l2(self._weight_decay),
                       use_bias=self._use_bias)(
            inputs)  # "One conv at the beginning (spatial size: 32x32)"

        # Add wide residual blocks
        block_fn = self._wide_basic
        conv2 = self._layer(block_fn, n_input_plane=n_stages[0], n_output_plane=n_stages[1],
                            count=n, stride=(1, 1))(conv1)
        conv3 = self._layer(block_fn, n_input_plane=n_stages[1], n_output_plane=n_stages[2],
                            count=n, stride=(2, 2))(conv2)
        conv4 = self._layer(block_fn, n_input_plane=n_stages[2], n_output_plane=n_stages[3],
                            count=n, stride=(2, 2))(conv3)
        batch_norm = BatchNormalization(axis=self._channel_axis)(conv4)
        relu = Activation("relu")(batch_norm)

        # Classifier block
        pool = AveragePooling2D(pool_size=(8, 8), strides=(1, 1), padding="same")(relu)
        flatten = Flatten()(pool)
        predictions_g = Dense(units=2, kernel_initializer=self._weight_init,
                              use_bias=self._use_bias,
                              kernel_regularizer=l2(self._weight_decay), activation="softmax",
                              name="pred_gender")(flatten)
        predictions_a = Dense(units=101, kernel_initializer=self._weight_init,
                              use_bias=self._use_bias,
                              kernel_regularizer=l2(self._weight_decay), activation="softmax",
                              name="pred_age")(flatten)
        model = Model(inputs=inputs, outputs=[predictions_g, predictions_a])

        return model


def main():
    model = WideResNet(64)()
    model.summary()


if __name__ == '__main__':
    main()
